#![allow(dead_code)]

use frenderer::camera::Camera;
use frenderer::types::*;
use frenderer::{Engine, Key, MousePos, Result};
use rand::prelude::*;
use std::rc::Rc;

const DT: f64 = 1.0 / 60.0;
const G: f32 = 15.0;

struct Ball {
    trf: Similarity3,
    model: Rc<frenderer::renderer::textured::Model>,
    vel: Vec3,
}
impl Ball {
    fn update(&mut self) {
        self.vel += Vec3::new(0.0, -G, 0.0) * DT as f32;
        self.trf.append_translation(self.vel * DT as f32);
        if self.trf.translation.y <= -400.0 {
            self.vel = Vec3::zero();
            let mut rng = rand::thread_rng();
            self.trf = Similarity3::new(
                Vec3::new(
                    rng.gen::<f32>() * 2.0 * W - W,
                    H,
                    rng.gen::<f32>() * 2.0 * W - W,
                ),
                Rotor3::identity(),
                rng.gen::<f32>() * SZ + MIN_SZ,
            );
        }
    }
    fn shape(&self) -> Sphere {
        Sphere {
            pos: self.trf.translation,
            r: self.trf.scale,
        }
    }
}
struct World {
    camera: Camera,
    camera_control: OrbitCamera,
    balls: Vec<Ball>,
    floor: Floor,
    contacts: Contacts,
}

// We could use fancy type stuff here or a more sophisticated type than
// usize if we wanted to be sure to prevent mixing up ball-ball and ball-floor contacts.

struct Contact {
    o1: usize,
    o2: usize,
    mtv: Vec3,
}

#[derive(Default)]
struct Contacts {
    ball_floor: Vec<Contact>,
    ball_ball: Vec<Contact>,
}

impl Contacts {
    fn new() -> Self {
        Self {
            ball_floor: vec![],
            ball_ball: vec![],
        }
    }
    fn sort(&mut self) {
        self.ball_floor
            .sort_unstable_by(|a, b| b.mtv.mag_sq().partial_cmp(&a.mtv.mag_sq()).unwrap());
        self.ball_ball
            .sort_unstable_by(|a, b| b.mtv.mag_sq().partial_cmp(&a.mtv.mag_sq()).unwrap());
    }
    fn clear(&mut self) {
        self.ball_floor.clear();
        self.ball_ball.clear();
    }
    fn push_ball_ball(&mut self, c: Contact) {
        self.ball_ball.push(c);
    }
    fn push_ball_floor(&mut self, c: Contact) {
        self.ball_floor.push(c);
    }
}

struct Sphere {
    pos: Vec3,
    r: f32,
}
struct Plane {
    // A normal, has to be a unit vector
    n: Vec3,
    // And a distance of how far along the normal this is
    d: f32,
}
struct Floor {
    trf: Similarity3,
    model: Rc<frenderer::renderer::textured::Model>,
}
impl Floor {
    fn shape(&self) -> Plane {
        Plane {
            n: self.trf.rotation * Vec3::unit_y(),
            d: self.trf.translation.y + self.trf.scale,
        }
    }
}
impl World {
    fn gather_contacts(&mut self) {
        self.contacts.clear();
        // GATHER CONTACTS HERE
        self.contacts.sort();
    }
    fn resolve_contacts(&mut self) {
        // DO RESTITUTION HERE
    }
}
const COLLISION_STEPS: usize = 3;
impl frenderer::World for World {
    fn update(&mut self, input: &frenderer::Input, _assets: &mut frenderer::assets::Assets) {
        let tilt_xy = input.key_axis(Key::Left, Key::Right) * 1.0 * DT as f32;
        let tilt_yz = input.key_axis(Key::Up, Key::Down) * 1.0 * DT as f32;
        for b in self.balls.iter_mut() {
            b.update();
        }
        self.floor
            .trf
            .append_rotation(Rotor3::from_rotation_xy(tilt_xy) * Rotor3::from_rotation_yz(tilt_yz));
        self.camera_control.update(input);
        self.camera_control.update_camera(&mut self.camera);
        for _step in 0..COLLISION_STEPS {
            self.gather_contacts();
            self.resolve_contacts();
        }
    }
    fn render(
        &mut self,
        _a: &mut frenderer::assets::Assets,
        rs: &mut frenderer::renderer::RenderState,
    ) {
        rs.set_camera(self.camera);
        for (bi, ball) in self.balls.iter().enumerate() {
            rs.render_textured(ball.model.clone(), ball.trf, bi);
        }
        rs.render_textured(self.floor.model.clone(), self.floor.trf, self.balls.len());
    }
}

const SZ: f32 = 5.0;
const MIN_SZ: f32 = 1.0;
const W: f32 = 200.0;
const H: f32 = 200.0;

fn main() -> Result<()> {
    frenderer::color_eyre::install()?;

    let mut engine: Engine = Engine::new(Default::default(), DT);

    let camera = Camera::look_at(
        Vec3::new(0., 0., 80.),
        Vec3::new(0., 0., 0.),
        Vec3::new(0., 1., 0.),
    );

    let floor_tex = engine.load_texture(std::path::Path::new("content/cube-diffuse.jpg"))?;
    let floor_meshes = engine.load_textured(std::path::Path::new("content/floor.obj"))?;
    let floor = engine.create_textured_model(floor_meshes, vec![floor_tex]);
    let marble_tex = engine.load_texture(std::path::Path::new("content/sphere-diffuse.jpg"))?;
    let marble = engine.load_textured(std::path::Path::new("content/sphere.obj"))?;
    let model = engine.create_textured_model(marble, vec![marble_tex]);
    let mut rng = rand::thread_rng();
    let world = World {
        camera,
        camera_control: OrbitCamera::new(),
        contacts: Contacts::default(),
        balls: (0..256)
            .map(|_| Ball {
                model: model.clone(),
                trf: Similarity3::new(
                    Vec3::new(
                        rng.gen::<f32>() * 2.0 * W - W,
                        rng.gen::<f32>() * H,
                        rng.gen::<f32>() * 2.0 * W - W,
                    ),
                    Rotor3::identity(),
                    rng.gen::<f32>() * SZ + MIN_SZ,
                ),
                vel: Vec3::zero(),
            })
            .collect(),
        floor: Floor {
            trf: Similarity3::new(Vec3::new(0.0, -25.0, 0.0), Rotor3::identity(), 10.0),
            model: floor,
        },
    };
    engine.play(world)
}

pub struct OrbitCamera {
    pub pitch: f32,
    pub yaw: f32,
    pub distance: f32,
}
impl OrbitCamera {
    fn new() -> Self {
        Self {
            pitch: 0.0,
            yaw: 0.0,
            distance: 400.0,
        }
    }
    fn update(&mut self, events: &frenderer::Input) {
        let MousePos { x: dx, y: dy } = events.mouse_delta();
        self.pitch += (DT * dy) as f32 / 10.0;
        self.pitch = self.pitch.clamp(-PI / 4.0, PI / 4.0);

        self.yaw += (DT * dx) as f32 / 10.0;
        self.yaw = self.yaw.clamp(-PI / 4.0, PI / 4.0);
        self.distance += events.key_axis(Key::W, Key::S) * 10.0 * DT as f32;
    }
    fn update_camera(&self, c: &mut Camera) {
        // The camera should point at the player (you could transform
        // this point to make it point at the player's head or center,
        // or at point in front of the player somewhere, instead of
        // their feet)
        let at = Vec3::zero();
        // And rotated around the player's position and offset backwards
        let camera_rot = Rotor3::from_euler_angles(0.0, self.pitch, self.yaw);
        let offset = camera_rot * Vec3::new(0.0, 0.0, -self.distance);
        let eye = offset;
        // To be fancy, we'd want to make the camera's eye an object
        // in the world whose rotation is locked to point towards the
        // player, and whose distance from the player is locked, and
        // so on---so we'd have player OR camera movements apply
        // accelerations to the camera which could be "beaten" by
        // collision.
        *c = Camera::look_at(eye, at, Vec3::unit_y());
    }
}

/// Are s1 and s2 touching?
fn touching_sphere_sphere(s1: &Sphere, s2: &Sphere) -> bool {
    // Is the (squared) distance between the centers less than the
    // (squared) sum of the radii?
    s2.pos.dot(s1.pos) <= (s1.r + s2.r).powi(2)
}
/// What's the offset I'd need to push s1 and s2 out of each other?
fn disp_sphere_sphere(s1: &Sphere, s2: &Sphere) -> Option<Vec3> {
    let offset = s2.pos - s1.pos;
    let distance = offset.mag();
    if distance < s1.r + s2.r {
        // Make sure we don't divide by 0
        let distance = if distance == 0.0 { 1.0 } else { distance };
        // How much combined radius is "left over"?
        let disp_mag = (s1.r + s2.r) - distance;
        // Normalize offset and multiply by the amount to push
        Some(offset * (disp_mag / distance))
    } else {
        None
    }
}
fn touching_sphere_plane(s: &Sphere, p: &Plane) -> bool {
    // Find the distance of the sphere's center to the plane
    (s.pos.dot(p.n) - p.d).abs() <= s.r
}
fn disp_sphere_plane(s: &Sphere, p: &Plane) -> Option<Vec3> {
    // Find the distance of the sphere's center to the plane
    let dist = s.pos.dot(p.n) - p.d;
    if dist.abs() <= s.r {
        // If we offset from the sphere position opposite the normal,
        // we'll end up hitting the plane at `dist` units away.  So
        // the displacement is just the plane's normal * (r-dist).
        Some(p.n * (s.r - dist))
    } else {
        None
    }
}
